# -*- coding: utf-8 -*-
#
# Copyright 2018-2019- Swiss Data Science Center (SDSC)
# A partnership between École Polytechnique Fédérale de Lausanne (EPFL) and
# Eidgenössische Technische Hochschule Zürich (ETHZ).
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Define models for Common Workflow Language."""

from ._ascwl import CWLClass, ascwl
from .command_line_tool import CommandLineTool
from .workflow import Workflow

WORKFLOW_STEP_RUN_TYPES = (CommandLineTool, Workflow)

__all__ = ('CommandLineTool', 'CWLClass', 'Workflow', 'ascwl')
